-----------------------------------------------------------------------------
-- |
-- Module      :  Main
-- Project     :  HBC
-- Copyright   :  (c) Hal Daume III
-- License     :  Open
-- 
-- Maintainer  :  Hal Daume III <me@hal3.name>
-- Stability   :  stable
-- Portability :  portable
--
-- Executable
--
-----------------------------------------------------------------------------

module Main where

import System
import System.IO
import Data.Char
import Data.List
import Data.Maybe
import Control.Monad
import Data.IORef
import qualified Data.Map as M

import Code
import Simulate
import Decl
import Parse
import Data
import Core
import MkCore
import qualified Conjugacy
import qualified UnLift
import qualified Math
import Marginalize(marginalizeAll)
import Type
import Data
import Control.Monad
import Control.Monad.State
import Util
import qualified CodeOpt

import qualified GenC(run)


data DumpSpec = DumpLast | DumpBest | DumpNth Int

data SimArgs = Define    String SimType
             | Collapse  String
             | LoadD     FilePath String String [String] [Char]
             | LoadM     FilePath String String String
             | File      FilePath
             | Verbosity Verbosity
             | Iter      Int
             | Dump      DumpSpec [String]
             | DumpCore
             | Make

programName = "hbc"

outputTypes = [(".c", GenC.run)]

usage msg = do
  mapM_ (hPutStrLn stderr) $
    ["usage: " ++ programName ++ " simulate [options] <HIER FILE>"
    ,"   or: " ++ programName ++ " compile  [options] <HIER FILE> <OUTPUT FILE>"
    ,""
    ,"where:"
    ,""
    ,"  <HIER FILE> is either the name of a file containing a hierarchical"
    ,"  model, or '-' for stdin and (for compilation) <OUTPUT FILE> is the"
    ,"  name of the desired output file for the compiled code (the extension"
    ,"  of the output file is used to determine the language)."
    ,""
    ,"simulation options are of the form:"
    ,"  --define <VAR> <VALUE>       to define a variable (which therefore won't"
    ,"                                 be sampled)"
    ,"  --collapse <VAR>             collapse (marginalize out/rao-blackwellize)"
    ,"                                 the given variable"
    ,"  --loadD <FILE> <VAR>* ';'    to load discrete data from <FILE>; it is"
    ,"                                 assumed to be in numeric form; the <VAR>"
    ,"                                 args should be: (1) the variable to store"
    ,"                                 the results in; (2) the variable to store"
    ,"                                 the highest read value in; (3..N) the"
    ,"                                 variables to store the dimensions (see"
    ,"                                 the user guide for details/examples)"
    ,"  --loadM <FILE> <V> <V> <V>   to load continuous data from <FILE>; the"
    ,"                                 first <VAR> is the matrix, the second"
    ,"                                 is the first dimension (number of lines),"
    ,"                                 the third is the second dimension (number"
    ,"                                 of values per line)"
    ,"  --dump <SPEC> <VAR>* ';'     dump variables in list according to <SPEC>:"
    ,"                                 last -- dump after the final iteration"
    ,"                                 best -- dump the sample with highest ll"
    ,"                                 #    -- dump every # samples"
    ,"                                 all  -- dump all samples (equiv to '1')"
    ,"  --[verbosity]                where [verbosity] is one of:"
    ,"                                 silent, quiet, verbose, trace, supertrace"
    ,"  --iter <NUM>                 where [NUM] is how many iterations to run"
    ,"  --dumpcore                   dump internal representations during startup"
    ,""
    ,"compilation options are of the form:"
    ,"  --define, --loadD, --loadM, --iter, --dumpcore, --collapse as in simulation"
    ,"  --make                       build executable output (only for .c)"
    ,""
    ,"currently supported output formats are: " ++ unwords (map fst outputTypes)
    ]
  
  when (not $ null msg) $ hPutStrLn stderr $ "\n" ++ msg
  exitWith (ExitFailure 1)


readValue x = 
  case readsPrec 0 x of
    [(v,"")] -> II v
    _ -> case readsPrec 0 x of
           [(v,"")] -> RR v
           _ -> error ("cannot parse value: " ++ x)

parseArgsSim ("--silent":xs) = Verbosity Silent : parseArgsSim xs
parseArgsSim ("--quiet":xs) = Verbosity Quiet : parseArgsSim xs
parseArgsSim ("--verbose":xs) = Verbosity Verbose : parseArgsSim xs
parseArgsSim ("--trace":xs) = Verbosity Trace : parseArgsSim xs
parseArgsSim ("--supertrace":xs) = Verbosity SuperTrace : parseArgsSim xs
parseArgsSim ("--iter":x:xs) = (Iter (read x)) : parseArgsSim xs
parseArgsSim ("--dumpcore":xs) = DumpCore : parseArgsSim xs
parseArgsSim ("--define":var:val:xs) = (Define var (readValue val)) : parseArgsSim xs
parseArgsSim ("--collapse":var:xs) = (Collapse var) : parseArgsSim xs
parseArgsSim ("--loadD":fn:var:high:rest) = 
  let dims = takeWhile (/=";") rest
      num  = length dims
      xs   = drop num rest
      seps = case num of
               0 -> ""
               1 -> ""
               2 -> "\n"
               3 -> "\n\t"
               _ -> error ("cannot read data with more than 3 dimensions")
  in  (LoadD fn var high dims seps) : parseArgsSim (drop 1 xs)
parseArgsSim ("--loadM":fn:var:n:d:xs) = (LoadM fn var n d) : parseArgsSim xs
parseArgsSim ("--dump":spec:rest) =
  let vars = takeWhile (/=";") rest
      xs   = drop (length vars+1) rest
      sp   = case spec of
               "last" -> DumpLast
               "best" -> DumpBest
               "all"  -> DumpNth 1
               n      -> (read n::Int) `seq` DumpNth (read n)
  in  (Dump sp vars) : parseArgsSim xs
parseArgsSim [fn] | not ("--" `isPrefixOf` fn) = [File fn]
parseArgsSim (x:xs) = error ("invalid argument beginning at: " ++ x)
parseArgsSim [] = error ("not enough arguments")

makeSampWarnings f = 
  when (all (\x -> case x of { Comment _ -> True ; _ -> False }) (body f)) $ do
    hPutStrLn stderr $ "Warning: no generated function body for " ++ name f
    mapM_ (\ (Comment s) -> hPutStrLn stderr $ "  " ++ s) (body f)
    hPutStrLn stderr ""

dumpInternal str z = do
  putStr $ "\n\n************************************* " ++ str ++ " *************************************\n\n"
  putStrLn $ show z
  putStrLn ""

buildCore dump coll mod = do
  when dump $ dumpInternal "mod" mod

  let (core,defType0) = filterTypeDefinitions $ mkCore mod
  when dump $ dumpInternal "core" core
  when dump $ dumpInternal "defType0" defType0

  let core' = nameSPVars $ UnLift.simplify core
  when dump $ dumpInternal "core'" core'

  let prior = Math.simplify $ coreToPrior core'
  when dump $ dumpInternal "prior" prior

  let lik   = removeSubF $ makeLikFunction prior (getGlobals core') (getTypeMap defType0 core')
  when dump $ dumpInternal "lik" lik

  let init  = removeSub $ [makeInitializeFunction prior d (getTypeMap defType0 prior) | d <- decls prior]
  when dump $ dumpInternal "init" init

  let conj  = fixOffsets $ nameSPVars $ Math.simplify $ UnLift.simplify $ Conjugacy.simplify core
  when dump $ dumpInternal "conj" conj

  let typ   = getTypeMap defType0 conj
  when dump $ dumpInternal "typ"  typ

  let (typ', marg, margIds, upd, warn) = marginalizeAll conj typ coll
  mapM_ (\s -> hPutStrLn stderr ("Warning: " ++ s)) warn

  when dump $ dumpInternal "marg" marg
  when dump $ dumpInternal "typ'" typ'
  when dump $ dumpInternal "upd"  upd

  let samp  = [f | d <- decls marg
                 , let f = makeSampleFunction d typ' upd
                 , not (("post_" ++ drop 9 (name f)) `elem` margIds)
                 ]
  when dump $ dumpInternal "samp" samp
  mapM_ makeSampWarnings samp

  let samp' = CodeOpt.opt $ removeSub $ removeID samp
  when dump $ dumpInternal "samp'" samp'

  let init' = map (makeMarginalInitialization typ' samp') margIds ++
              [f | f <- init
                 , not (("post_" ++ drop 11 (name f)) `elem` margIds)]
  when dump $ dumpInternal "init'" init'

  return (typ', lik, init', samp')

getVars vt vars = M.filterWithKey (\k _ -> k `elem` vars) vt

printDump ll t = do
  putStrLn $ "== SAMPLE\t" ++ show ll
  flip mapM_ (M.assocs t) $ \ (var, val) ->
    showSimType val >>= \showVal ->
    putStrLn $ "== " ++ var ++ "\t" ++ showVal ""

makeDump _  (Dump _ []) _ _ = return ()
makeDump di (Dump (DumpNth n) vars) st ll = do
  (cnt,d) <- readIORef di
  writeIORef di (cnt+1,d)
  when ((cnt `mod` n) == (n-1)) $ printDump ll (getVars st vars)
makeDump di (Dump DumpLast vars) st ll = do
  (cnt,_) <- readIORef di
  writeIORef di (cnt, Just (ll, getVars st vars))
makeDump di (Dump DumpBest vars) st ll = do
  (cnt,d) <- readIORef di
  case d of
    Just (ll0,_) | ll0 > ll -> return ()
    _                       -> writeIORef di (cnt, Just (ll, getVars st vars))

simulate :: [SimArgs] -> IO ()
simulate [] = usage ""
simulate args = do
  fn  <- case last args of { File fn -> return fn ; _ -> usage "missing hierarchical filename" }
  (mod,direct) <- loadHier fn
  simulate' ((parseArgsSim (direct ++ [""])) ++ init args) mod

simulate' args mod = do
  lD  <- flip mapM loadD $ \ (LoadD fn var high dims seps) -> do
           (varV, highV, dimsV) <- loadDiscrete seps fn
           return ((var,varV) : (high,highV) : zip dims dimsV)
  lM  <- flip mapM loadM $ \ (LoadM fn var n d) -> do
           (varV, nV, dV) <- loadMatrix fn
           return [(var,varV), (n,nV), (d,dV)]
  let decls = vals ++ concat lD ++ concat lM
  (_, likF, initF, sampF) <- buildCore dumpc coll mod
  
  dumpInfo <- newIORef (0, Nothing)
  runM verb $ simAll numi likF initF sampF decls (makeDump dumpInfo dump)

  case dump of 
    Dump _ [] -> return ()
    Dump (DumpNth _) _ -> return ()
    _ -> do
      (_,d) <- readIORef dumpInfo
      case d of
        Nothing -> return ()
        Just (ll,v) -> printDump ll v
  return ()
  where
    verb  = last (Verbose : mapMaybe (\x -> case x of { Verbosity v -> Just v ; _ -> Nothing }) args)
    vals  = mapMaybe (\x -> case x of { Define var val -> Just (var,val) ; _ -> Nothing }) args
    loadD = mapMaybe (\x -> case x of { f@(LoadD {}) -> Just f ; _ -> Nothing }) args
    loadM = mapMaybe (\x -> case x of { f@(LoadM {}) -> Just f ; _ -> Nothing }) args
    numi  = last (100 : mapMaybe (\x -> case x of { Iter v -> Just v ; _ -> Nothing }) args)
    dump  = last ((Dump DumpLast []) : mapMaybe (\x -> case x of { d@(Dump {}) -> Just d ; _ -> Nothing }) args)
    dumpc = not $ null $ filter (\x -> case x of { DumpCore -> True ; _ -> False }) args
    coll  = mapMaybe (\x -> case x of { Collapse var -> Just var ; _ -> Nothing }) args

compile args = 
  case find ((`isSuffixOf` fout) . fst) outputTypes of
    Nothing -> usage "unknown output file type"
    Just (suff, run) -> do
      (mod, direct) <- loadHier fin
      compile' ((parseArgsCom (direct ++ ["",""])) ++ args) suff run mod
  where
    [File fin, File fout] = drop (length args - 2) args

compile' args suff run mod = do
      (typ, likF, initF, sampF) <- buildCore dumpc coll mod
      let defaults = (vals, loadD, loadM, numi)
      run defaults dumpc fout typ likF initF sampF
      when (make && suff==".c") $ do
        let cmd = "gcc -lm samplib.c stats.c " ++ fout ++ " -o " ++ (init fout ++ "out")
        hPutStrLn stderr $ "running: "++ cmd
        system cmd
        return ()
  where
    [File fin, File fout] = drop (length args - 2) args

    vals  = mapMaybe (\x -> case x of { Define var val -> Just (var,val) ; _ -> Nothing }) args
    loadD = mapMaybe (\x -> case x of { f@(LoadD fn var high dims seps) -> Just (var,(fn,high,dims,seps)) ; _ -> Nothing }) args
    loadM = mapMaybe (\x -> case x of { f@(LoadM fn var n d) -> Just (var,(fn,n,d)) ; _ -> Nothing }) args
    numi  = last (100 : mapMaybe (\x -> case x of { Iter v -> Just v ; _ -> Nothing }) args)
    dumpc = not $ null $ filter (\x -> case x of { DumpCore -> True ; _ -> False }) args
    make  = not $ null $ filter (\x -> case x of { Make     -> True ; _ -> False }) args
    coll  = mapMaybe (\x -> case x of { Collapse var -> Just var ; _ -> Nothing }) args

parseArgsCom [fin,fout] = [File fin, File fout]
parseArgsCom ("--iter":x:xs) = (Iter (read x)) : parseArgsCom xs
parseArgsCom ("--dumpcore":xs) = DumpCore : parseArgsCom xs
parseArgsCom ("--define":var:val:xs) = (Define var (readValue val)) : parseArgsCom xs
parseArgsCom ("--collapse":var:xs) = (Collapse var) : parseArgsCom xs
parseArgsCom ("--loadD":fn:var:high:rest) = 
  let dims = takeWhile (/=";") rest
      num  = length dims
      xs   = drop num rest
      seps = case num of
               0 -> ""
               1 -> ""
               2 -> "\n"
               3 -> "\n\t"
               _ -> error ("cannot read data with more than 3 dimensions")
  in  (LoadD fn var high dims seps) : parseArgsCom (drop 1 xs)
parseArgsCom ("--loadM":fn:var:n:d:xs) = (LoadM fn var n d) : parseArgsCom xs
parseArgsCom ("--make":xs) = Make : parseArgsCom xs
parseArgsCom (x:_) = error ("invalid arguments beginning at: " ++ show x)

runMain ("simulate":xs) = simulate (parseArgsSim xs)
runMain ("compile" :xs) = compile  (parseArgsCom xs)
runMain _ = usage ""

main = getArgs >>= runMain


-- runMain $ words "simulate --dump best z mu pi ; --loadM X x N dim --define K 2 mix_gauss.hier"
-- runMain $ words "simulate --dump best z ; --loadD testW w V D N ; --define K 2 LDA.hier"
-- runMain $ words "simulate --dump best z ; --loadD testW w V N   ; --define K 2 HMM.hier"
-- runMain $ words "simulate --dump best z ; --loadD testW w V D N ; --define K 2 LDA-HMM.hier"


-- runMain $ words "compile --make --loadM X x N dim       --define K 2 mix_gauss.hier mix_gauss.c"
-- runMain $ words "compile --make --loadD testW w V D N ; --define K 2 LDA.hier       LDA.c"
-- runMain $ words "compile --make --loadD testW w V N   ; --define K 2 HMM.hier       HMM.c"
-- runMain $ words "compile --make --loadD testW w V D N ; --define K 2 LDA-HMM.hier   LDA-HMM.c"

